/*-
 * Copyright (C)2013 @BABOLO http://www.babolo.ru/
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#ifndef lint
static const char copyright[] = "\
@(#)Copyright (C)2013 @BABOLO http://www.babolo.ru/\n\
@(#)All rights reserved.\n";
static const char rcsid[] = "$Id: anypost.c,v 1.8 2014/03/27 17:36:27 babolo Exp $";
#endif /* not lint */

#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <sys/event.h>
#include <sys/time.h>
#include <sysexits.h>
#include <sys/uio.h>
#include <strings.h>
#include <string.h>
#include <unistd.h>
#include <stdlib.h>
#include <limits.h>
#include <netdb.h>
#include <stdio.h>
#include <errno.h>
#include <math.h>
#include <err.h>
#include <babolo/BLINflag.h>
#include <babolo/parser.h>
#include <mife.h>

#define CRV 256
#define WRV 10
#define HTVER 2

typedef enum
{   anypost_none = 0
,   anypost_http = 1
,   anypost_udp  = 2
} anypost_sh;

typedef struct {
    BLIN_flag  flags;
#       define ANYPOST_SEEREQ  0x800000
#       define ANYPOST_NOCONTL 0x400000
#       define ANYPOST_NOHDRS  0x200000
#       define ANYPOST_NOBODY  0x100000
#       define ANYPOST_SEEANS  0x080000
#       define ANYPOST_CHUNKD  0x040000
#       define ANYPOST_ADDNUL  0x020000
#       define ANYPOST_CRLFEL  0x000800
#       define ANYPOST_LFLFEL  0x000400
#       define ANYPOST_CHUNKY  0x000100
    anypost_sh sh;
    mife_time  itime;
    mife_time  otime;
    char      *host;
    char      *port;
    char      *user;
    char      *pass;
    char      *get;
    char      *areq;
    int        proto;
} anypost_cf;

typedef struct {
    char            *f;
    char            *a;
} anypost_header;

static char *empty = "";
static char *get = "POST ";
static char *end = "\r\n";
static char *httpver[] =
{ ""
, " HTTP/1.0\r\nHost: "
, " HTTP/1.1\r\nHost: "
};

static void
usage(int ex) {
    fprintf( stderr
           , "anypost " VERS "  " DATE "\n"
             "    -a req    - request instead from stdin\n"
             "    -B        - display on stdout answer body only\n"
             "    -c        - no Content-Length: header\n"
             "    -C        - chunk assembly\n"
             "    -d        - duplicate request to stderr\n"
             "    -D        - duplicate answer to stderr\n"
             "    -E        - use environment variable instead of optarg\n"
             "    -d        - duplicate request to stderr\n"
             "    -f header - add header\n"
             "    -F file   - add headers from file\n"
             "    -h ver    - HTTP version, 1 - HTTP/1.0, 2 - HTTP/1.1\n"
             "    -H        - display on stdout answer headers only\n"
             "    -n        - add nul to request\n"
             "    -q        - quiet\n"
             "    -t time   - timeouts after init, sec\n"
             "    -T time   - init timeout\n"
             "    -v        - verbose\n"
             "    -z ....   - recursive opts without environment substitute\n"
             "    -Z ....   - recursive opts with environment substitute\n"
           )
    ;
    exit(ex);
}

static void
retest(u_int32_t flags, char *o) {
    for (; *o; o++) {
        if  (o[1] == ':') {
            printf("%c\t1\n", *(o++));
        } else {
            printf("%c\t0\n", *o);
    }   }
//  if  () printf("\\.\n");
}

#define NOPT  0
#define OPT0  1
#define OPT1  2

static int
retext(babolo_opt *opt) {
    int c, ex = EX_OK;
    babolo_opt *o;

    for (o = opt; o; o = o->next) {
        for (c = 0; c < 256; ++c) {
            if  (o->opts[c]) {
                if  (c < ' ') {
                    printf("\\%03o\t%u\n", c, o->opts[c] - 1);
                } else {
                    printf("%c\t%u\n", c, o->opts[c] - 1);
    }   }   }   }
    return(ex);
}

static char *
rediropt(anypost_cf *options, babolo_opt **opt, int redirenv) {
    ifBLIN_FLAG = options->flags;
    char *e, *o;

    if  (!(o = babolo_getoptarg(opt, 0))) {
        ifBLIN_QW0("No argument");
        usage(EX_USAGE);
    } else if ((redirenv > 0) && (!(o = getenv(e = o)))) {
        ifBLIN_QX0("No ENV %s", e);
        usage(EX_USAGE);
    }
    return(o);
}

int
main(int argc, char *argv[], char **envp) {
    int c, s, w, redirenv, error, crv, wrv, ex = EX_OK;
    struct addrinfo *hint, *res, *res0;
    mife_descriptor *mdi, *mdh;
    size_t respl, wrto, wrin;
    anypost_cf *options;
    char *flar, *o, *p;
    anypost_header *cr;
    BLIN_flag subflags;
    struct iovec *buf;
    ssize_t postsize;
    babolo_parm *bp;
    babolo_opt *opt;
    double tout;
    void *resp;

    blin_ctl(BLIN_CTL_FSET, BLIN_BIT0);
    if  (!(options = calloc(1, sizeof(*options)))) {
        ifBLIN_QX0("Malloc failed #1");
        errno = ENOMEM;
        ex = EX_OSERR;
        goto out;
    }
    options->proto = HTVER;
    options->flags = BLIN_BIT0;
    if  (!(cr = calloc(CRV, sizeof(*cr)))) {
        ifBLIN_QX0("Malloc failed");
        errno = ENOMEM;
        ex = EX_OSERR;
        goto out;
    }
    opt = NULL;
    if  (babolo_openopt(Bpars_NONU, &opt, argc, argv, flar = "a:BCcDdEF:f:Hh:nqt:T:vz:Z:?=:")) {
        ifBLIN_QW0("openopt");
        ex = EX_SOFTWARE;
        goto out;
    }
    crv = 0;
    for (redirenv = 0; (c = babolo_getopt(&opt)) != 0; redirenv--) {
        if  (redirenv < 0) redirenv = 0;
        switch (c) {
        case 'a': options->areq = rediropt(options, &opt, redirenv);                             break;
        case 'B': options->flags |= ANYPOST_NOHDRS;                                              break;
        case 'C': options->flags |= ANYPOST_CHUNKD;                                              break;
        case 'c': options->flags |= ANYPOST_NOCONTL;                                             break;
        case 'D': options->flags |= ANYPOST_SEEANS;                                              break;
        case 'd': options->flags |= ANYPOST_SEEREQ;                                              break;
        case 'E': redirenv = 2;                                                                  break;
        case 'F': if  (crv >= CRV) {
                ;     ifBLIN_QX0("Too many headers, %d max", CRV);
                ;     usage(EX_USAGE);
                ; }
                ; cr[crv++].f = rediropt(options, &opt, redirenv);
                ; break;
        case 'f': if  (crv >= CRV) {
                ;     ifBLIN_QX0("Too many headers, %d max", CRV);
                ;     usage(EX_USAGE);
                ; }
                ; cr[crv++].a = rediropt(options, &opt, redirenv);
                ; break;
        case 'H': options->flags |= ANYPOST_NOBODY;                                              break;
        case 'h': o = rediropt(options, &opt, redirenv);
                ; options->proto = strtoul(o, NULL, 0);
                ; if  (options->proto > HTVER) {
                ;     ifBLIN_QX1("htver %d > %d", options->proto, HTVER);
                ;     options->proto = HTVER;
                ; }
                ; break;
        case 'n': options->flags |= ANYPOST_ADDNUL;                                              break;
        case 'q': BLIN_QUIET(options->flags);                                                    break;
        case 't': o = rediropt(options, &opt, redirenv);
                ; tout = atof(o);
                ; options->otime.tv_nsec = lrint(floor(1000000000 * modf(tout, &tout)));
                ; options->otime.tv_sec = lrint(tout);
                ; ifBLIN_QX2("Other timeout %"BLIN_D".%09ld\n", options->otime.tv_sec, options->otime.tv_nsec);
                ; break;
        case 'T': o = rediropt(options, &opt, redirenv);
                ; tout = atof(o);
                ; options->itime.tv_nsec = lrint(floor(1000000000 * modf(tout, &tout)));
                ; options->itime.tv_sec = lrint(tout);
                ; ifBLIN_QX2("Init timeout %"BLIN_D".%09ld\n", options->itime.tv_sec, options->itime.tv_nsec);
                ; break;
        case 'v': BLIN_VERBOSE(options->flags);                                                  break;
        case 'z': o = rediropt(options, &opt, redirenv);
                ; bp = babolo_getparm( Bpars_NOEN | Bpars_NOAL | Bpars_NONU
                                     , (u_char**)&o
                                     , (u_char*)""
                                     , 0
                                     );
                ; babolo_openopt(0, &opt, bp->argc, (char **)bp->argv, NULL);
                ; break;
        case 'Z': o = rediropt(options, &opt, redirenv);
                ; bp = babolo_getparm(Bpars_NOAL | Bpars_NONU, (u_char**)&o, (u_char*)"", 0);
                ; babolo_openopt(0, &opt, bp->argc, (char **)bp->argv, NULL);
                ; break;
        case '=': if  (!(o = rediropt(options, &opt, redirenv))) {
                      ifBLIN_QW0("No -= arg");
                      ex = EX_USAGE;
                  } else {
                      switch (o[0] & 0xFF) {
                      case '0': retest('T', flar); break;
                      case '1': retext(opt);       break;
                  }   }
                ; goto out;
        case '?': usage(EX_OK);
        default : ifBLIN_QX0("Illegal flag -%c", c);
                ; usage(EX_USAGE);
    }   }
    if  (!(o = babolo_getarg(&opt))) {
        ifBLIN_QW0("URL absent");
        ex = EX_USAGE;
        goto out;
    }
    subflags = ((BLIN_GEN1 & options->flags) | (options->flags >> 1)) & BLIN_MASK;
    if  (!strncmp(o, "udp://", strlen("udp://"))) {
        options->sh = anypost_udp;
        o += strlen("udp");
    } else if (!strncmp(o, "http://", strlen("http://"))) {
        options->sh = anypost_http;
        o += strlen("http");
    } else {
        ifBLIN_QX0("URL=%s~: Illegal shema", o);
        errno = EINVAL;
        ex = EX_USAGE;
        goto out;
    }
    *o++ = '\0';
    o += strlen("//");
    if  ((p = strpbrk(o, "/@")) && *p == '@') {
        /* username */
        *p++ = '\0';
        options->user = o;
        if  ((o = index(o, ':'))) {
            *o++ = '\0';
            options->pass = o;
        }
        o = p;
    }
    options->host = o;
    if  ((p = strpbrk(o, "/:")) && *p == ':') {
        /* port */
        *p++ = '\0';
        options->port = p;
        o = p;
    }
    if  (o && (o = index(o, '/'))) *o++ = '\0';
    options->get = o ? o : empty;
    wrto = strlen(options->get) + ((options->flags & ANYPOST_ADDNUL) ? 1 : 0);
    if  (!options->port) options->port = "80";
    ifBLIN_QQ2(options->flags) {
        ifBLIN_QX2("sh=%d proto=%d\n", options->sh, options->proto);
        if  (options->user) ifBLIN_QX2("user=%s~\n", options->user);
        if  (options->pass) ifBLIN_QX2("pass=%s~\n", options->pass);
        if  (options->host) ifBLIN_QX2("host=%s~\n", options->host);
        if  (options->port) ifBLIN_QX2("port=%s~\n", options->port);
        if  (options->get) ifBLIN_QX2("get=%s~\n", options->get);
        for (c = 0; c < crv; ++c) {
            if  (cr[c].f) ifBLIN_QX2("f[%d]=%s~\n", c, cr[c].f);
            if  (cr[c].a) ifBLIN_QX2("a[%d]=%s~\n", c, cr[c].a);
    }   }
    if  (!(hint = calloc(1, sizeof(*hint)))) {
        ifBLIN_QX0("Malloc failed");
        errno = ENOMEM;
        ex = EX_OSERR;
        goto out;
    }
    hint->ai_family = PF_INET;
    switch (options->sh) {
    case anypost_http:
        hint->ai_socktype = SOCK_STREAM;
        hint->ai_protocol = IPPROTO_TCP;
        break;
    case anypost_udp:
        hint->ai_socktype = SOCK_DGRAM;
        hint->ai_protocol = IPPROTO_UDP;
        if  (  (options->flags & ANYPOST_SEEREQ)
            && (wrto != (wrin = write(fileno(stderr), options->get, wrto)))
            ) {
            ifBLIN_QW0("write debug %d <> %d", wrin, wrto);
            ex = EX_IOERR;
            goto out;
        }
        break;
    default:
        ifBLIN_QX0("%d sh", options->sh);
        errno = EPROTONOSUPPORT;
        ex = EX_PROTOCOL;
    }
    hint->ai_flags = AI_ADDRCONFIG;
    if  ((error = getaddrinfo(options->host, options->port, hint, &res))) {
        ifBLIN_QX0("%d=%s", error, gai_strerror(error));
        errno = ENETUNREACH;
        ex = EX_NOHOST;
        goto out;
    }
    for (s = -1, res0 = res; res; res = res->ai_next) {
        if  (0 > (s = socket(res->ai_family, res->ai_socktype, res->ai_protocol))) {
            ifBLIN_QW1( "socket=%d family=%d socktype=%d protocol=%d fail"
                      , s, res->ai_family, res->ai_socktype, res->ai_protocol
                      );
            continue;
        }
        switch (options->sh) {
        case anypost_http:
            if  (0 > (connect(s, res->ai_addr, res->ai_addrlen))) {
                ifBLIN_QW1("connect s=%d", s);
                close(s);
                s = -1;
                continue;
            }
            break;
        case anypost_udp:
            if  (0 > (wrin = sendto( s, options->get, wrto
                                   , MSG_NOSIGNAL | MSG_EOR | MSG_EOF
                                   , res->ai_addr, res->ai_addrlen
                )    )             ) {
                ifBLIN_QW1("sendto s=%d", s);
                close(s);
                s = -1;
                continue;
            }
            if  (wrto != wrin) {
                ifBLIN_QW0("sendto %d <> %d", wrin, wrto);
                ex = EX_IOERR;
                goto out;
            }
            break;
        default:
            ifBLIN_QX0("%d sh", options->sh);
            errno = EPROTONOSUPPORT;
            ex = EX_PROTOCOL;
        }
        break;
    }
    if  (s < 0) {
        ifBLIN_QX0("Not connected %s:%s", options->host, options->port);
        ex = EX_UNAVAILABLE;
        goto out;
    }
    freeaddrinfo(res0);
    if  (options->sh == anypost_udp) goto out;
    for (wrv = WRV, c = 0; c < crv; ++c) {
        if  (cr[c].f) ++wrv;
        if  (cr[c].a) wrv += 2;
    }
    ifBLIN_QX3("buf %d x %"BLIN_U"\n", wrv, BLIN_I(sizeof(*buf)));
    if  (!(buf = calloc(wrv, sizeof(*buf)))) {
        ifBLIN_QX0("Malloc failed");
        errno = ENOMEM;
        ex = EX_OSERR;
        goto out;
    }
    w = 0;
    buf[w].iov_base = get;
    buf[w++].iov_len = strlen(get);
    buf[w].iov_base = "/";
    buf[w++].iov_len = 1;
    buf[w].iov_base = options->get;
    buf[w++].iov_len = strlen(options->get);
    if  (options->proto) {
        buf[w].iov_base = httpver[options->proto];
        buf[w++].iov_len = strlen(httpver[options->proto]);
        buf[w].iov_base = options->host;
        buf[w++].iov_len = strlen(options->host);
        buf[w].iov_base = end;
        buf[w++].iov_len = strlen(end);
    }
    for (c = 0; c < crv; ++c) {
        if  (cr[c].f) {
            if  (!(mdi = mife_init(subflags))) {
                ifBLIN_QW0("mife_init");
                ex = EX_IOERR;
                goto out;
            }
            if  (0 > mife_ctl(mdi, MIFE_CTLFILE, cr[c].f)) {
                ifBLIN_QW0("mife_ctl -F=%s~", cr[c].f);
                ex = EX_IOERR;
                goto out;
            }
            if  (0 > (buf[w].iov_len = mife_read(mdi, 0, 0))) {
                ifBLIN_QW0("mife_read -F=%s~");
                ex = EX_IOERR;
                goto out;
            }
            if  (!(o = mife_get(mdi, 0))) {
                ifBLIN_QW0("mife_get -F=%s~");
                ex = EX_IOERR;
                goto out;
            }
            buf[w].iov_base = strndup(o, buf[w].iov_len);
            mife_fini(mdi);
            mdi = NULL;
            ++w;
        }
        if  (cr[c].a) {
            buf[w].iov_base = cr[c].a;
            buf[w++].iov_len = strlen(cr[c].a);
            buf[w].iov_base = end;
            buf[w++].iov_len = strlen(end);
    }   }
    if  (options->areq) {
        postsize = strlen(options->areq);
    } else {
        if  (!(mdi = mife_init(subflags))) {
            ifBLIN_QW0("mife_init");
            ex = EX_IOERR;
            goto out;
        }
        if  (0 > mife_ctl(mdi, MIFE_CTLFDSC, fileno(stdin))) {
            ifBLIN_QW0("mife_ctl stdin");
            ex = EX_IOERR;
            goto out;
        }
        if  (0 > (postsize = mife_read(mdi, 0, 0))) {
            ifBLIN_QW0("mife_read stdin");
            ex = EX_IOERR;
            goto out;
    }   }
    if  (!(options->flags & ANYPOST_NOCONTL)) {
        if  (0 > (buf[w].iov_len = asprintf((void*)&buf[w].iov_base, "Content-Length: %u", postsize))
            ) {
            ifBLIN_QW0("asprintf failed");
            errno = ENOMEM;
            ex = EX_OSERR;
            goto out;
        }
        buf[++w].iov_base = end;
        buf[w++].iov_len = strlen(end);
    }
    buf[w].iov_base = end;
    buf[w++].iov_len = strlen(end);
    buf[w].iov_len = postsize;
    if  (options->areq) {
        buf[w++].iov_base = options->areq;
    } else if (!(buf[w++].iov_base = mife_get(mdi, 0))) {
        ifBLIN_QW0("mife_get stdin");
        ex = EX_IOERR;
        goto out;
    }
    if  (w > wrv) {
        ifBLIN_QX0("w=%u over wrv=%u", w, wrv);
        errno = EDOOFUS;
        ex = EX_SOFTWARE;
        goto out;
    }
    for (wrto = 0, c = 0; c < w; ++c) {
        ifBLIN_QX2("[%d]=%s~\n", c, buf[c].iov_base);
        wrto += buf[c].iov_len;
    }
    if  ((options->flags & ANYPOST_SEEREQ) && (wrto != (wrin = writev(fileno(stderr), buf, w)))) {
        ifBLIN_QW0("writev debug %d <> %d", wrin, wrto);
        ex = EX_IOERR;
        goto out;
    }
    if  (wrto != (wrin = writev(s, buf, w))) {
        ifBLIN_QW0("writev %d <> %d", wrin, wrto);
        ex = EX_IOERR;
        goto out;
    }
    mife_fini(mdi);
    mdi = NULL;
    if  (!(mdh = mife_init(subflags))) {
        ifBLIN_QW0("mife_init");
        ex = EX_IOERR;
        goto out;
    }
    if  (0 > mife_ctl(mdh, MIFE_CTLFDSC, s)) {
        ifBLIN_QW0("mife_ctl socket");
        ex = EX_IOERR;
        goto out;
    }
    if  (0 > mife_ctl(mdh, MIFE_CTLITIM, &options->itime)) {
        ifBLIN_QW0("mife_ctl itime");
        ex = EX_IOERR;
        goto out;
    }
    if  (0 > mife_ctl(mdh, MIFE_CTLOTIM, &options->otime)) {
        ifBLIN_QW0("mife_ctl otime");
        ex = EX_IOERR;
        goto out;
    }
    ifBLIN_QX2( "  i=%u.%09lu o=%u.%09u\n"
              , mdh->itime.tv_sec
              , mdh->itime.tv_nsec
              , mdh->otime.tv_sec
              , mdh->otime.tv_nsec
              );
    if  (0 > (respl = mife_read(mdh, 0, 0))) {
        ifBLIN_QW0("mife_read socket");
        ex = EX_IOERR;
        goto out;
    }
    ifBLIN_QX4("read %"BLIN_D"\n", BLIN_I(respl));
    if  (!(resp = mife_get(mdh, 0))) {
        ifBLIN_QW0("mife_get socket");
        ex = EX_IOERR;
        goto out;
    }
    if  ((options->flags & ANYPOST_SEEANS) && (respl != mife_writ(fileno(stderr), resp, respl))) {
        ifBLIN_QW0("mife_writ stderr");
        ex = EX_IOERR;
        goto out;
    }
    if  (!(options->flags & (ANYPOST_NOHDRS | ANYPOST_NOBODY))) {
        if  (respl != mife_writ(fileno(stdout), resp, respl)) {
            ifBLIN_QW0("mife_writ");
            ex = EX_IOERR;
            goto out;
        }
    } else {
        o = strnstr(resp, "\n\r\n", respl);
        p = strnstr(resp, "\n\n", respl);
        if  (o) ifBLIN_QX4("CRLF %"BLIN_U"\n", BLIN_I((void*)o - resp));
        if  (p) ifBLIN_QX4("LFLF %"BLIN_U"\n", BLIN_I((void*)p - resp));
        if  (!p) {
            if  (o) {
                p = ++o + 2;
                options->flags |= ANYPOST_CRLFEL;
                ifBLIN_QX3("CRLF no LFLF %u + %u\n", (void*)o - resp, resp - (void*)p + respl);
            } else {
                ifBLIN_QX3("no CRLF no LFLF\n");
            }
        } else if (!o) {
            o = ++p;
            ++p;
            options->flags |= ANYPOST_LFLFEL;
            ifBLIN_QX3("LFLF no CRLF %u + %u\n", (void*)o - resp, resp - (void*)p + respl);
        } else if (p > o) {
            p = ++o + 2;
            options->flags |= ANYPOST_CRLFEL;
            ifBLIN_QX3("CRLF before LFLF %u + %u\n", (void*)o - resp, resp - (void*)p + respl);
        } else {
            o = ++p;
            ++p;
            options->flags |= ANYPOST_LFLFEL;
            ifBLIN_QX3("LFLF before CRLF %u + %u\n", (void*)o - resp, resp - (void*)p + respl);
        }
        if  (!(options->flags & ANYPOST_NOHDRS)) {
            ifBLIN_QX3("HDRS\n");
            if  (((void*)o - resp) != mife_writ(fileno(stdout), resp, (void*)o - resp)) {
                ifBLIN_QW0("mife_writ hdrs");
                ex = EX_IOERR;
                goto out;
            }
            if  (  !(options->flags & ANYPOST_NOBODY)
                && ((p - o) != mife_writ(fileno(stdout), o, p - o))
                ) {
                ifBLIN_QW0("mife_writ sepa");
                ex = EX_IOERR;
                goto out;
        }   }
        if  (!(options->flags & ANYPOST_NOBODY)) {
            ifBLIN_QX3("BODY\n");
            if  (options->flags & ANYPOST_CHUNKD) {
                ifBLIN_QX3("CHUNKD\n");
                for (o = resp; o < p; o = memchr(o, '\n', o - p) + 1) {
                    ifBLIN_QX3("%c%c%c%c%c%c%c%c\n", o[0], o[1], o[2], o[3], o[4], o[5], o[6], o[7]);
                    if  (  (o - p > strlen("Transfer-Encoding:"))
                        && !strncasecmp(o, "Transfer-Encoding:", strlen("Transfer-Encoding:"))
                        && ((void*)strnstr(o, "chunked", o - p) < (void*)memchr(o, '\n', o - p))
                        ) {
                        ifBLIN_QX3("Transfer-Encoding: chunked\n");
                        options->flags |= ANYPOST_CHUNKY;
                        break;
            }   }   }
            if  (!(options->flags & ANYPOST_CHUNKY)) {
                ifBLIN_QX3( "writ(%d, 0x%"BLIN_X", %"BLIN_U")\n"
                          , fileno(stdout)
                          , BLIN_I(p)
                          , BLIN_I(resp - (void*)p + respl)
                          );
                if  (  (resp - (void*)p + respl)
                    != mife_writ(fileno(stdout), p, resp - (void*)p + respl)
                    ) {
                    ifBLIN_QW0("mife_writ body");
                    ex = EX_IOERR;
                    goto out;
                }
            } else {
                size_t t;

                for (o = p; (void*)o < (void*)(resp + respl); o += t) {
                    t = strtoul(o, &o, 16);
                    ifBLIN_QX3("size %X(%d)\n", t, t);
                    if  (!t) break;
                    if  (options->flags & ANYPOST_CRLFEL) {
                        o = strnstr(o, "\r\n", resp - (void*)o + respl) + 2;
                    } else {
                        o = strnstr(o, "\n", resp - (void*)o + respl) + 1;
                    }
                    ifBLIN_QX3( "writ(%d, 0x%"BLIN_X", %"BLIN_U")\n"
                              , fileno(stdout)
                              , BLIN_I(o)
                              , BLIN_I(t)
                              );
                    if  (t != mife_writ(fileno(stdout), o, t)) {
                        ifBLIN_QW0("mife_writ chunk");
                        ex = EX_IOERR;
                        goto out;
    }   }   }   }   }
out:
    exit(ex);
}
